/*****************************************************************************

    Name: crefdb.c
    Description: fuzzy search library - search procedure
    
                 
    Author: Oleg Khovayko,  http://olegh.spedia.net

   ***************************************************************************

                          PUBLIC DOMAIN NOTICE
              National Center for Biotechnology Information

    This software/database is a "United States Government Work" under the
    terms of the United States Copyright Act.  It was written as part of    
    the author's official duties as a United States Government employee
    and thus cannot be copyrighted.  This software/database is freely
    available to the public for use. The National Library of Medicine and
    the U.S. Government have not placed any restriction on its use or
    reproduction.

    Although all reasonable efforts have been taken to ensure the accuracy
    and reliability of the software and data, the NLM and the U.S.
    Government do not and cannot warrant the performance or results that
    may be obtained by using this software or data. The NLM and the U.S.
    Government disclaim all warranties, express or implied, including
    warranties of performance, merchantability or fitness for any
    particular purpose.

    Please cite the author in any work or product based on this material.

   ***************************************************************************

    Entry Points:



*****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <fuzzy.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <math.h>

/*--------------------------------------------------------------------------*/

int fuzzy_find(FDB_HANDLE *h, const char *term, FUZZY_RESULT *out, int limit) {
  TRIADE_QTY qty_buf[5 * MAXTERMLEN];
  int n_triades, last_triade_no;
  int   i, len;
  FUZZY_MERGE *cur, *prev;
  Int4 cur_term_no;
  float curweight;
  Uint4 cmp_template_hi, cmp_template_lo;

  // int summary_xrs;
  // int really_xrs;

  //  summary_xrs = really_xrs = 0; // Global statistic

  FUZZY_MERGE merge_rollist[5 * MAXTERMLEN];
  FUZZY_MERGE *rollist_ptr;

  FUZZY_FOUND *found_list;
  int found_listN;

  n_triades = calculate_noncanonic_triades(term, qty_buf, 2);

  if(limit == 0 || n_triades == 0 || fuzzy_init(h) == 0) 
     return 0; // Not inited or empty expression


                // -- Create merge lists
  for(last_triade_no = i = 0; 
      i < n_triades; i++) { // init elements for merge rollist
    FUZZY_MERGE *rlo = &merge_rollist[last_triade_no];
    rlo->triade      = qty_buf[i].triade_r50;
    XRS_ENTRY   *xre = &h->statmas_entry[rlo->triade];
    if(xre->size > 0) {
      last_triade_no++;
      rlo->weight      = xre->global_weight * count_to_lw[qty_buf[i].qty];
      rlo->ptr         = &h->xrs_table[xre->offset];
      rlo->qty         = xre->size; 
      //  summary_xrs += xre->size;
    }
    //  char tr[4];
    // r50_to_triade(rlo->triade, tr);
    // printf("ROLADD: [%s] %d (%f)\n", tr, rlo->qty, rlo->weight);
  }


  if(last_triade_no == 0) // User enters triades, which we never used
    return 0;

  for(i = 0; i < last_triade_no; i++) { // linking merge rollist
    merge_rollist[i].next = &merge_rollist[i + 1];
    if(merge_rollist[i].qty == 0)
      fprintf(stderr, "ERROR: merge_rollist[i].qty == 0\n");
  } // for

  last_triade_no--;
  merge_rollist[last_triade_no].next = NULL;

  rollist_ptr = merge_rollist;
  
  found_listN = 0; 
  found_list = (FUZZY_FOUND*)malloc(limit * sizeof(FUZZY_FOUND));
  if(found_list == NULL) {
    fprintf(stderr, "Can't malloc(%d) for found_list\n", limit * sizeof(FUZZY_FOUND));
       return 0;
  } 

  do {

    // Stage 1
    // Find minimum cur_term_no and it's weight

    cmp_template_lo = cmp_template_hi = (~0) >> 1;

    for(prev = rollist_ptr; prev != NULL;  prev = prev->next) { 
      Uint4 xrs = *prev->ptr;
      if(xrs <= cmp_template_hi) { 
        if(xrs < cmp_template_lo) { // new minimum term_no
          cmp_template_hi = xrs |  QTY_MASK;
          cmp_template_lo = xrs & ~QTY_MASK;
          cur = prev;
          i   = 0;
        } else 
          i++;
      } // 
    }

    if(i == 0 && cur->qty > 1) {
      curweight = cur->weight * GET_TERM_LW(*(cur->ptr)++);
      cur->qty--;
    } else { // --------------

      if(cur->qty <= 1) 
        cur = rollist_ptr;
    
      curweight = 0; // reset accumulator

      do {

        if(*(cur->ptr) <= cmp_template_hi) { // This is lowest (our) term_no
          curweight += cur->weight * GET_TERM_LW(*(cur->ptr)++);
          // cur->qty_lw[lw]--;
          // really_xrs++;
          if((--cur->qty) == 0) {
            // No more values for this element - remove element and step to next
            if(cur == rollist_ptr) {
              rollist_ptr = cur->next;
            } else { 
              cur = prev->next = cur->next;
              continue;
            }
          }
        } // if(*(cur->ptr) < cmp_template) -- This is lowest (our) term_no

        prev = cur; cur = cur->next;

      } while(cur);

    } // ------------

    // Stage 2
    // Insert cur_term_no into found_list (weight in the curweight)

    if(found_listN < limit || curweight > found_list[0].weight) { // Need insert
      cur_term_no = GET_TERM_NO(cmp_template_hi);

      curweight *= h->termlist_entry[cur_term_no].len_corr;

      if(found_listN == 0) {
        found_listN = 1; i = 0;

      } else {
        if(found_listN >= limit && curweight <= found_list[0].weight)
          goto no_insert;
          
        // BinSearch for position for insert
        int l, u;
        l = 0; u = found_listN - 1;
        do {
          i = (l + u) >> 1;
          if(curweight > found_list[i].weight) 
            l = i + 1;
          else
            u = i - 1;
        } while(u >= l);

        i = l;

        if(found_listN < limit) {
          len = found_listN - i; found_listN++;
          if(len > 0)
            memmove(&found_list[i + 1], &found_list[i], len * sizeof(FUZZY_FOUND));
        
        } else {
          if((--i) > 0)
            memmove(&found_list[0], &found_list[1], i * sizeof(FUZZY_FOUND));
        }
      }
      found_list[i].term_no = cur_term_no;
      found_list[i].weight  = curweight;
    } //if(...) -- Need insert


    no_insert:;

    
  } while(rollist_ptr);

  
  for(i = 0; i < found_listN; i++) {
    Int4 term_pos;
    int j = found_listN - i - 1;
    out[i].rate     = found_list[j].weight;
    out[i].term_no  = cur_term_no = found_list[j].term_no;
    term_pos        = h->termlist_entry[cur_term_no].offset;
    out[i].term     = h->termlist + term_pos;
    out[i].term_len = h->termlist_entry[cur_term_no + 1].offset - term_pos - 1;
  }

  free(found_list);
  return found_listN;

}


/*==================================================================================*/

