/*****************************************************************************

    Name: crefdb.c
    Description: Create DB-structures for fuzzy search
    
                 
    Author: Oleg Khovayko,  http://olegh.spedia.net

   ***************************************************************************

                          PUBLIC DOMAIN NOTICE
              National Center for Biotechnology Information

    This software/database is a "United States Government Work" under the
    terms of the United States Copyright Act.  It was written as part of    
    the author's official duties as a United States Government employee
    and thus cannot be copyrighted.  This software/database is freely
    available to the public for use. The National Library of Medicine and
    the U.S. Government have not placed any restriction on its use or
    reproduction.

    Although all reasonable efforts have been taken to ensure the accuracy
    and reliability of the software and data, the NLM and the U.S.
    Government do not and cannot warrant the performance or results that
    may be obtained by using this software or data. The NLM and the U.S.
    Government disclaim all warranties, express or implied, including
    warranties of performance, merchantability or fitness for any
    particular purpose.

    Please cite the author in any work or product based on this material.

   ***************************************************************************

    Entry Points:

    input: 

  ./termlist.txt - File with vocabulary

  <STDIN> - File with frequency user requests. Used for collect
statistic and optimize data structures for spedup search

   output:
   termlist.txt.offset - offsets to lines in the file ./termlist.txt
   ./triades.xrs       - Cross reference file (size: up to 4*termlist.txt)
   ./triades.xrs.entry - entry table for ./triades.xrs (size: 1024000 bytes)


*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <fuzzy.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <math.h>

#define VALBUFSIZ 2048

typedef struct {
    Uint4  qty;
    Uint4  siz;
    Uint4  pos;
    float  global_weight;
    Uint4  *valbuf;
    Uint2  valbufqty;
    Uint2  valbufsiz;
    R50  triade;
    Int4  sortqty[QTY_MASK + 1];
} STAT;

static STAT *statmas;
static XRS_ENTRY *statmas_entry;
static Uint4  *xrs_table;

/*--------------------------------------------------------------------------*/


int stat_compare(const void *p1, const void *p2) {
  STAT *a = (STAT*)p1;
  STAT *b = (STAT*)p2;
  float res = b->qty * b->global_weight 
            - a->qty * a->global_weight;

  return res == 0? 0 : (res > 0? 1 : -1);

}
/*--------------------------------------------------------------------------*/

FILE *fopen_termlist() {
  FILE *fterms;
    if((fterms = fopen("termlist.txt", "r")) == NULL) {
      perror("can't open file termlist");
      exit(1);
    }
    return fterms;
}

/*--------------------------------------------------------------------------*/

int fetch_term(FILE *fterms, char *buf) {
  *buf = 0;

  if(fgets(buf, MAXTERMLEN - 1, fterms)) {
        char *ctrlp = strchr(buf, 'P' & 037);
        if(ctrlp)
          *ctrlp = 0;
        return 1;
  }

  return 0;
}

/*--------------------------------------------------------------------------*/

void download(STAT *stat) {
  if(stat == NULL)
    return;

  if(stat->valbufqty) {
    memcpy(&xrs_table[stat->pos], stat->valbuf, stat->valbufqty * sizeof(Uint4));
    stat->pos += stat->valbufqty;
    stat->siz -= stat->valbufqty;
    stat->valbufqty = 0;
  }

  if(stat->siz == 0 && stat->valbuf != NULL) {
    free(stat->valbuf); stat->valbuf = NULL;
  }

  if(stat->siz < 0) {
    fprintf(stderr, "Overflow XRS entry!\n");
  }
}
/*--------------------------------------------------------------------------*/
float  calculate_len_corr(TRIADE_QTY *qty_buf, int n) {
  float sum = 0;
  while(--n >= 0) {
    int lw = count_to_lw[qty_buf[n].qty];
    sum += lw * lw * statmas_entry[qty_buf[n].triade_r50].global_weight;
  }
  return 1 / sqrt(sum);
}
/*--------------------------------------------------------------------------*/
int main(int agc, char **argv) {
    int i, n;
    char term[MAXTERMLEN];
    TRIADE_QTY qty_buf[5 * MAXTERMLEN];
    TERM_ENTRY entry;

    FILE *fterms, *foffset, *fstat;

    if((statmas = (STAT*)calloc(MAXR50, sizeof(STAT)))        == NULL ||
     (statmas_entry = (XRS_ENTRY*)calloc(MAXR50, sizeof(XRS_ENTRY))) == NULL) {
        perror("Can't get buffer for statmas");
        exit(1);
    }

    for(i = 0; i < MAXR50; i++) 
        statmas[i].triade = i;


    printf("Analyze file termlist.txt and calculate triade_qty\n");

    fterms = fopen_termlist();

    int summary_terms = 0;

    while(fetch_term(fterms, term)) {
      // putw((int)ftell(fterms), foffset);
        n = calculate_noncanonic_triades(term, qty_buf, 1);
        for(i = 0; i < n; i++) {
            R50 triade = qty_buf[i].triade_r50;
            statmas[triade].qty += qty_buf[i].qty;
            statmas[triade].siz++;
            statmas[triade].sortqty[count_to_lw[qty_buf[i].qty]]++;
        }
        summary_terms++;

    } /* while(gets(term)) */

    printf("Collect statistic for stdin\n");

    while(fgets(term, MAXTERMLEN - 1, stdin)) {
        n = calculate_noncanonic_triades(term, qty_buf, 1);
        for(i = 0; i < n; i++) {
            statmas[qty_buf[i].triade_r50].qty += qty_buf[i].qty;
        }
    } /* while(gets(term)) */


    printf("Sorting\n");

    int summary_triades = 0;
    int nonempty_triades = 0;
    for(i = 0; i < MAXR50; i++) { 
      if(statmas[i].siz) {
        nonempty_triades++;
        summary_triades += statmas[i].qty;
      }
    }
    
    for(i = 0; i < MAXR50; i++) 
        statmas[i].global_weight = statmas[i].siz? 
          log(summary_triades / (statmas[i].siz + 10)) : 0;
 
    qsort(statmas, MAXR50, sizeof(STAT), stat_compare);

    printf("Print result\n");

    if((fstat = fopen("freqstat.txt", "w")) == NULL) {
      perror("Can't open statistic file for write");
      exit(1);
    }

    fprintf(fstat,
     "Triades: all=%d; active=%d; summary_triades=%d; summary_terms=%d\n", 
           MAXR50, nonempty_triades, summary_triades, summary_terms);

    int pos_offset = 0;
    for(i = 0; i < MAXR50; i++) {
      r50_to_triade(statmas[i].triade, term);
      fprintf(fstat, "%s\t%7d (%2d%%)\t%7d (%2d%%);", 
             term, 
             statmas[i].qty, statmas[i].qty * 100 / summary_triades, 
             statmas[i].siz, statmas[i].siz * 100 / summary_terms);

      if(statmas[i].siz) {
        R50 triade = statmas[i].triade;
        statmas_entry[triade].pos = i;
        statmas_entry[triade].offset = pos_offset;
        statmas_entry[triade].size = statmas[i].siz;
        statmas_entry[triade].global_weight = statmas[i].global_weight;
      }

      fprintf(fstat, "\tPos=%d; gw=%f\n", pos_offset, statmas[i].global_weight);

      statmas[i].pos = pos_offset;
      pos_offset += statmas[i].siz;

    }

    fclose(fstat);

    printf("Try mmap()\n");
    rewind(fterms);
 
    int fd = open("triades.xrs", O_RDWR | O_CREAT | O_TRUNC, 0644);
    lseek(fd, pos_offset * sizeof(Uint4) - 1, SEEK_SET);
    write(fd, "", 1);
    xrs_table = (Uint4*)mmap(NULL, pos_offset * sizeof(Uint4), 
                     PROT_READ | PROT_WRITE, MAP_SHARED | MAP_NORESERVE,
                     fd, 0);

    printf("mmap() OK. Try fill XRS-file\n");

    foffset = fopen("termlist.txt.entry", "wb");
    entry.offset   = 0;
    entry.len_corr = 0;

    fwrite(&entry, sizeof(TERM_ENTRY), 1, foffset); // for fictive term #0
    int term_no = 0; /* terms 1..N */
    while(fetch_term(fterms, term)) {
        term_no += (1 << BITS_FOR_LW);
        if((term_no & (077777 << BITS_FOR_LW)) == 0) {
          printf("Processed terms: %d\r", term_no >> BITS_FOR_LW);
          fflush(stdout);
        }

        n = calculate_noncanonic_triades(term, qty_buf, 1);

        entry.len_corr = calculate_len_corr(qty_buf, n);

        fwrite(&entry, sizeof(TERM_ENTRY), 1, foffset); // for  term 
        entry.offset = ftell(fterms);

        for(i = 0; i < n; i++) {
          R50 triade = qty_buf[i].triade_r50;
          STAT *stat = &statmas[statmas_entry[triade].pos];
          Uint4 val = term_no | count_to_lw[qty_buf[i].qty];
          if(stat->triade != triade) {
            fprintf(stderr, "Invalid stat_entry\n");
          }

          if(stat->valbuf == NULL) {
            if(stat->siz < VALBUFSIZ)
              stat->valbufsiz = stat->siz;
            else
              stat->valbufsiz = VALBUFSIZ;

            if((stat->valbuf = (Uint4*)malloc(stat->valbufsiz * sizeof(Uint4))) == NULL) {
              perror("No memory for valbuf");
              exit(1);
            }
          } /* if(stat->valbuf == NULL) */

          stat->valbuf[stat->valbufqty++] = val;
          if(stat->valbufqty >= stat->valbufsiz) {
            download(stat);
          }

          if(stat->siz < 0) {
            fprintf(stderr, "Overflow XRS entry!\n");
          }
        }
    } /* while(gets(term)) */

    entry.len_corr = 0;
    fwrite(&entry, sizeof(TERM_ENTRY), 1, foffset); // for fictive last term
    fclose(foffset);

    printf("Processed terms: %d\n", term_no >> BITS_FOR_LW);

    printf("\nDownload buffers and check pointers\n");

    for(i = 0; i < MAXR50; i++) {
      download(&statmas[i]);
    }
    for(i = 0; i < MAXR50; i++) {
      if(!statmas_entry[i].offset)
        continue;
      if(statmas_entry[i].offset + statmas_entry[i].size != statmas[statmas_entry[i].pos].pos)
        fprintf(stderr, "Error in the download: %d + %d != %d\n",
        statmas_entry[i].offset, statmas_entry[i].size, statmas[statmas_entry[i].pos].pos);

    }
    
    printf("Check OK\n");
   
    // No-SORT NOW!!!

#if 0

    for(i = 0; i < MAXR50; i++) {
      STAT *stat = &statmas[i];
      if(statmas_entry[stat->triade].size > 1) { // Need sorting 
        stat->siz = statmas_entry[stat->triade].size;
        stat->pos = statmas_entry[stat->triade].offset;
        char triade_text[4];
        // r50_to_triade(stat->triade, triade_text);
        // fprintf(stderr, "Sort #%d [%s]: (%d)\r", i, triade_text, stat->siz);

        Uint4 *buf = (Uint4 *)malloc(stat->siz * sizeof(Uint4));
        Uint4 *masptr[QTY_MASK + 1];

        if(buf == NULL) {
          perror("Can't get memory for sorting\n");
          continue;
        }
        int j;
        Uint4 *ptr = buf;
        for(j = n = 0; j <= QTY_MASK; j++) 
          n += stat->sortqty[j];

        if(n != stat->siz) 
          fprintf(stderr, "\nNot EQ!!! n == %d;  stat->siz == %d \n", n, stat->siz);
        
        for(j = QTY_MASK; j >= 0; j--) {
          masptr[j] = ptr;
          ptr += stat->sortqty[j];
        }

        for(j = 0; j < stat->siz; j++) {
          Uint4 val = xrs_table[stat->pos + j];
          *masptr[val & QTY_MASK] = val;
          masptr[val & QTY_MASK]++;
        } 
       
        memcpy(&xrs_table[stat->pos], buf, stat->siz * sizeof(Uint4));
        free(buf);
      } /* if(statmas_entry[stat->triade].size > 1) */
    } /* for(i = 0; i < MAXR50; i++)  */

#endif


    printf("munmap(); processed %d terms\n", term_no >> BITS_FOR_LW);
    /* munmap((void*)xrs_table, pos_offset * sizeof(Uint4)); */
    close(fd);
    fclose(fterms);

    if((fd = open("triades.xrs.entry", O_WRONLY  | O_CREAT | O_TRUNC, 0644)) > 0
       && write(fd, statmas_entry, MAXR50 * sizeof(XRS_ENTRY)) > 0)
       close(fd);
    else
      perror("Can't write file: triades.xrs.entry");

    printf("finished OK\n");

    return 0;
}

